<?php

namespace App\Http\Controllers;

use App\Config;
use App\Customer;
use App\DefaultMessage;
use App\Invoice;
use App\Item as Article;
use App\Sale;
use App\SalesOrder;
use App\Session;
use App\User;
use Carbon\Carbon;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Schema;
use Mike42\Escpos\PrintConnectors\WindowsPrintConnector;
use Mike42\Escpos\Printer;


class InvoiceController extends Controller
{

    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
    }

    public function index(Request $request)
    {
        $user = $request->user();
        if ($user->is_admin) {
            $sales = Sale::with(['user', 'items', 'service'])->latest()->limit(100)->get();
        } else {
            $sales = $user->session->sales()->with(['user', 'items', 'service'])->latest()->get();
        }

        return response()->json($sales);
    }

    public function store(Request $request): JsonResponse
    {
        $this->validate($request, [
            'items' => 'required|min:3',
            // 'variants' => 'required|array|min:0',
            // 'modifiers' => 'required|array|min:0',
            // 'taxes' => 'required|array|min:0'
        ], [
            'items.min' => 'List of articles cannot be empty.'
        ]);
        $sale = null;
        $total = 0;
        DB::transaction(function () use ($request, &$sale, &$total) {
            $user = $request->user();
            // $service_id = $request->input('service', 0);
            $session = $user->session;
            $customer = $request->input('customer');
            // $discount = $request->input('discount', []);

            $paid = $request->input('paid');
            $whId = $user->department->warehouse->id;
            $payments = $request->payments;

            $inputs = is_string($request->items) ? json_encode($request->items) : $request->items;
            $item_ids = array_map(function ($item) {
                return $item['id'];
            }, $inputs);
            $items = $user->department->warehouse->items()->whereIn('item_id', $item_ids)->with(['category', 'price_lists'])->get();

            $savable_items = [];
            $next_sales_ref = get_next_ref('sale');

            foreach ($items as $item) {
                $curr_item_index = array_search($item['id'], array_column($inputs, "id"));
                $curr_item = $inputs[$curr_item_index];
                if ($item->category->tracked && $curr_item['quantity'] > $item->pivot->quantity) {
                    return response()->json([
                        'status' => 'error',
                        'message' => str_replace(
                            ['{requested}', '{item}', '{quantity}'],
                            [$curr_item['quantity'], $curr_item['name'], $item->pivot->quantity],
                            "The requested quantity for {item} ({requested}) exceeds stock level ({quantity}).")
                    ], 409);
                }
                $discounts = [];
                $unit_price = null;
                if (!$unit_price || $unit_price <= 0) {
                    $unit_price = $item->pivot->selling_price ?: $item->unit_price;
                }

                $total += $unit_price * (int)$curr_item['quantity'];
                $savable_items[] = [
                    'item_id' => $item->id,
                    'quantity' => $curr_item['quantity'],
                    'user_id' => $user->id,
                    'session_id' => $session->id,
                    // 'service_id' => $service_id,
                    'unit_price' => $unit_price,
                    'cost_price' => $item->purchase_price,
                    'created_at' => date('Y-m-d H:i:s'),
                    'updated_at' => date('Y-m-d H:i:s')
                ];
            }
            $inv_ids = [];
            $payable_discount_aware = $total - $discount;

            // No customer selected and payment is incomplete
            if ($payable_discount_aware > $paid && !$customer) {
                return response()->json([
                    'status' => 'error',
                    'message' => "Please, select or create a customer for incomplete payments."
                ], 302);
            }

            foreach ($savable_items as $inv) {
                // Save invoice items and get their ids ($inv_ids) for later
                $inv_ids[] = Invoice::insertGetId($inv);
            }

            if (count($inv_ids)) {
                $sale = new Sale();
                $sale->user_id = $user->id;
                $sale->code = $next_sales_ref;
                $sale->total_price = $total;
                $sale->session_id = $session->id;
                $sale->service_id = $service_id;
                $sale->customer_id = $customer ? $customer->id : null;
                $sale->discount = $discount;
                $sale->paid = $paid;
                $sale->balance = ($paid - $payable_discount_aware) > 0 ? ($paid - $payable_discount_aware) : 0;
                $sale->debt = ($paid - $payable_discount_aware) > 0 ? 0 : abs($paid - $payable_discount_aware);
                if (Schema::hasColumn(app(Sale::class)->getTable(), 'received')) {
                    $sale->received = $request->input('received', 0);
                }
                $sale->save();
                // Link invoice items to sales
                $sale->items()->attach($inv_ids);

                // Link payments
                $payment_keys = [];
                $payment_values = [];
                array_map(function ($payment) use ($user, &$payment_keys, &$payment_values) {
                    $payment_keys[] = $payment['id'];
                    $payment_values[] = ['amount' => $payment['amount'], 'user_id' => $user->id];
                }, $payments);
                $sale->payments()->attach(array_combine($payment_keys, $payment_values));

                // Deduct items from stock
                $whCtrl = new WarehouseController();
                $message = DefaultMessage::where('module', 'item_sale')->first();
                foreach ($savable_items as $idx => $item) {
                    $curr_item_index = array_search($item['item_id'], array_column($items->toArray(), "id"));
                    $curr_item = $items[$curr_item_index];

                    $req = new Request();
                    $req->setMethod('POST');
                    $req->request->add([
                        'quantity' => $item['quantity'],
                        'type' => 'dec',
                        'ref' => $next_sales_ref,
                        'reason' => str_replace(
                            ['{customer}', '{created}', '{item}', '{quantity}', '{item_ref}', '{sales_ref}'],
                            [$customer ? $customer->name : __('sale.customer.anonymous'), substr($sale->created_at, 0, 10), $curr_item['name'], $item['quantity'], $curr_item['code'], $next_sales_ref],
                            $message->message),
                        'date' => date('Y-m-d'),
                        'sales_price' => $item['unit_price'],
                        'warehouse' => $whId
                    ]);
                    $whCtrl->adjust_stock($item['item_id'], $req);
                }

                // Mark sales_order as done if passed
                if ($request->anyFilled('sales_order')) {
                    $sales_order = SalesOrder::find($request->input('sales_order'));
                    if ($sales_order) {
                        $soCtrl = new SalesOrderController();
                        $soCtrl->markComplete($sales_order->id);
                    }
                }
            }

            /*return response()->json([
                'status' => 'success',
                'data' => $sale // Sale::with('items', 'session', 'service', 'cashier')->find($sale->id),
            ]);*/
            /* return response()->json([
                 'status' => 'error',
                 'message' => 'No items selected.'
             ], 302);*/
        });

        return response()->json([
            'status' => 'success',
            'data' => $sale // Sale::with('items', 'session', 'service', 'cashier')->find($sale->id),
        ]);
    }

    // Not being used
    public function getInvoices(Request $request)
    {
        $user = $request->user();
        $inv = [];
        if ($user->is_admin) {
            $inv = Invoice::with(['user' => function ($query) use ($user) {
                $query->with('department');
            }]);
        } else {
            $inv = Session::find($user->session->id)->with(['user' => function ($query) use ($user) {
                $query->with('department');
            }, 'invoices']);
        }
        $inv = $inv->with(['item'])->latest()->limit(100)->get();

        return $inv->toArray();
    }

    public function getCashierSessionInvoices(Request $request)
    {
        $user = $request->user();
        $invoices = Sale::with(['items', 'customer', 'service', 'payments'])->withCount('items')
            ->where('session_id', $user->session->id)->latest()->paginate(10);

        return response()->json([
            'status' => 'success',
            'data' => $invoices
        ]);
    }

    private function getCreatorNames($sessions)
    {
        $names = "";
        foreach (Session::whereIn('id', $sessions)->get() as $index => $session) {
            if (!str_contains($names, $session->user->name)) {
                if ($index > 0) $names .= ", ";
                $names .= $session->user->name;
            }
        }
        return $names;
    }
}
