<?php

namespace App\Http\Controllers;

use App\Category;
use App\Department;
use App\Item;
use App\ItemPriceList;
use App\ItemVariant;
use App\Stock;
use App\User;
use App\VariantOption;
use App\Variant;
use App\Warehouse;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Schema;
use Illuminate\Support\Facades\Storage;
use Illuminate\Support\Str;
use Illuminate\Validation\ValidationException;

class ItemController extends Controller
{
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
        //
    }

    public function getItems()
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user', 'category.department', 'price_lists', 'variants', 'modifiers', 'taxes')->active()->latest()->get()
        ]);
    }

    public function mobileSaleItems(): JsonResponse
    {
        $data = Category::with(['items' => function (Builder $query) {
            $query->with('price_lists', 'variants', 'modifiers', 'taxes');
        }]);
        return response()->json([
            'status' => 'success',
            'data' => $data
        ]);
    }

    public function getActiveItems()
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::active()->get()
        ]);
    }

    public function getItemCategories()
    {
        return response()->json([
            'status' => 'success',
            'data' => Category::with(['items' => function ($query) {
                return $query->with('user', 'category.department', 'price_lists')->active();
            }])->get()
        ]);
    }

    public function getPOSItems()
    {
        $data = User::find(auth()->id())->department->warehouse->items()->with(['category', 'price_lists'])->active('items.state')->get();
        // or auth()->user()->department->warehouse->items()->with(['category', 'price_lists'])->get();
        /*$dept = auth()->user()->department;
        if (!$dept) return response()->json(['status' => 'success', 'data' => []]);

        $wh = $dept->warehouse;
        $stocks = Warehouse::with('items.category', 'items.price_lists')->find($wh->id);
        $data = [];
        foreach ($stocks->items->toArray() as $stock) {
            if (!isset($data[$stock['category']['name']])) {
                $data[$stock['category']['name']] = $stock["category"];
                $data[$stock['category']['name']]['items'] = [];
            }

            $stock['unit_price'] = $stock["pivot"]["selling_price"];
            $stock['quantity'] = $stock["pivot"]["quantity"];
            $data[$stock['category']['name']]['items'][] = $stock;
        }*/
        return response()->json([
            'status' => 'success',
            'data' => $data/*array_values($data)*/
        ]);
    }

    public function getItemsUnitTypes()
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::select('unit_type as value')->distinct()->get()
        ]);
    }

    public function nextRef(): ?string
    {
        return get_next_ref('item');
    }

    public function getItem($id)
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user')->find($id)
        ]);
    }

    /**
     * @throws ValidationException
     */
    public function createItem(Request $request)
    {
        $this->validate($request, [
            'name' => 'required',
            'purchase_price' => 'nullable|numeric',
            'unit_type' => 'nullable|string',
            'category' => 'required',
            'colour' => 'required_if:visual_type,colour',
            'image' => 'required_if:visual_type,image'
        ], [
            'name.required' => 'Please, provide a name.',
            'unit_type.required' => 'Please Specify a unit type',
        ]);

        // Log::info(json_encode($request->all()));

        $item = new Item();
        $item->name = $request->input('name');
        $item->unit_type = $request->input('unit_type');
        $item->category_id = $request->input('category');

        // @todo check setting unit_price
//        if ($request->hasAny('unit_price') && count($request->input('unit_price'))) {
//            $item->unit_price = $request->input('unit_price', 0.0);
//        }

        if ($request->hasAny('item_type')) {
            $item->item_type = $request->input('item_type');
        }

        if (Schema::hasColumn('items', 'supplier_id')) {
            $item->supplier_id = $request->input('supplier');
        }

        if (Schema::hasColumn('items', 'code')) {
            $item->code = get_next_ref('item');
        }

        if ($request->filled('reorder_level')) {
            $item->reorder_level = $request->input('reorder_level', 0);
        }

        if (Schema::hasColumn('items', 'optimal_order')) {
            $item->optimal_order = $request->input('optimal_order');
        }

        if (Schema::hasColumn('items', 'sellable')) {
            $item->sellable = $request->input('sellable', false);
        }

        if (Schema::hasColumn('items', 'tracked')) {
            $item->tracked = $request->input('tracked');
        }

        if (Schema::hasColumn('items', 'sell_by')) {
            $item->sell_by = $request->input('sell_by');
        }

        $item->description = $request->input('description');
        $item->purchase_price = $request->input('purchase_price');
        $item->barcode = $request->input('barcode');
        $item->user_id = auth()->id();

        if (Schema::hasColumn('items', 'visual_type')) {
            $item->visual_type = $request->input('visual_type', 'colour');
        }

        if ($request->get('visual_type', 'colour') == 'colour') {
            $item->colour = $request->input('colour', Item::defaultColour);
        } elseif ($request->has('image')) {
            $imageName = $this->storeBase64Image($request);
            $item->image = '/storage/' . $imageName;
        } else {
            $item->colour = Item::defaultColour;
        }

        $item->save();

        // Add variants
        $variants = $request->input('variants');
        if ($request->has('variants') && count($variants)) {
            $item_variants = (new VariantController())->save_variant($variants, $item->id);
        }

        // Add modifiers
        $modifiers = $request->input('modifiers', []);
        if ($modifiers && is_array($modifiers) && count($modifiers)) {
            $item->modifiers()->attach(array_map(function ($m) {
                return $m['id'];
            }, $modifiers));
        }

        // Add taxes
        $taxes = $request->input('taxes', []);
        if ($taxes && is_array($taxes) && count($taxes)) {
            $item->taxes()->attach(array_map(function ($tax) {
                return $tax['id'];
            }, $taxes));
        }

        $price_lists = array_filter($request->input('price_lists', []), function ($field) {
            return !!$field['value'];
        });

        if (count($price_lists)) {
            $sync_data = [];
            foreach ($price_lists as $price_list) {
                $sync_data[$price_list['id']] = ['price' => $price_list['value']];
            }
            $item->price_lists()->sync($sync_data);
        }

        $department_prices = array_filter($request->input('departments'), function ($field) {
            return !!$field['selling_price'];
        });
        if (count($department_prices)) {
            $wh_ids = [];
            foreach ($department_prices as $department_price) {
                // Add stock item to department's warehouse
                $wh_id = Department::find($department_price['id'])->warehouse_id;
                if ($wh_id) {
                    Warehouse::find($wh_id)->items()->attach($item->id, ['selling_price' => $department_price['selling_price']]);
                }
                $wh_ids[] = $wh_id;
            }
            // Add stock to MAGAZINE WAREHOUSE (ID: 1)
            $default_wh = Warehouse::whereDefault(true)->first();
            if ($default_wh && !in_array($default_wh->id, $wh_ids)) {
                // @todo $item->unit_price may not have been set. Try using default selling price for item in wh
                $default_wh->items()->attach($item->id, ['selling_price' => $item->unit_price]);
            }
        }

        return response()->json([
            'status' => 'success',
            'data' => Item::with('user', 'category.department', 'price_lists', 'variants', 'modifiers', 'taxes')->find($item->id)
        ]);
    }

    /**
     * @param Request $request
     * @return string
     */
    public function storeBase64Image(Request $request, $key = 'image'): string
    {
        $image_64 = $request->input($key);
        $extension = explode('/', explode(':', substr($image_64, 0, strpos($image_64, ';')))[1])[1];   // .jpg .png .pdf
        $replace = substr($image_64, 0, strpos($image_64, ',') + 1);

        $image = str_replace($replace, '', $image_64);
        $image = str_replace(' ', '+', $image);
        $imageName = config('app.uploads.item_images') . Str::random(10) . '.' . $extension;
        Storage::disk('public')->put($imageName, base64_decode($image));
        return $imageName;
    }

    /**
     * @throws ValidationException
     */
    public function updateItem($id, Request $request): JsonResponse
    {
        $this->validate($request, [
            'name' => 'required',
            'purchase_price' => 'nullable|numeric',
            'unit_type' => 'nullable|string',
            'category' => 'required',
            'colour' => 'required_if:visual_type,colour',
            'image' => 'required_if:visual_type,image'
        ], [
            'name.required' => 'Please, provide a name name.',
            'unit_type.required' => 'Please Specify a unit type',
        ]);
        DB::transaction(function () use ($request, $id) {
            $item = Item::with('user', 'category.department')->find($id);
            $item->name = $request->input('name');
            $item->unit_price = $request->input('unit_price');
            $item->unit_type = $request->input('unit_type');
            $item->item_type = $request->input('item_type');
            $item->category_id = $request->input('category');
            if (Schema::hasColumn('items', 'supplier_id')) {
                $item->supplier_id = $request->input('supplier');
            }
            $item->description = $request->input('description');
            $item->purchase_price = $request->input('purchase_price');
            $item->barcode = $request->input('barcode');

            if ($request->filled('reorder_level')) {
                $item->reorder_level = $request->input('reorder_level', 0);
            }

            if (Schema::hasColumn('items', 'optimal_order')) {
                $item->optimal_order = $request->input('optimal_order');
            }

            if (Schema::hasColumn('items', 'sellable')) {
                $item->sellable = $request->input('sellable', false);
            }

            if (Schema::hasColumn('items', 'sell_by')) {
                $item->sell_by = $request->input('sell_by');
            }

            if (Schema::hasColumn('items', 'tracked')) {
                $item->tracked = $request->input('tracked');
            }

            $item->description = $request->input('description');
            $item->purchase_price = $request->input('purchase_price');
            $item->barcode = $request->input('barcode');
            $item->user_id = auth()->id();

            if ($request->get('visual_type') == 'colour' && $request->input('colour') != $item->colour) {
                $item->colour = $request->input('colour', null);

                $this->removeImageIfExist($item->image);
                $item->image = null;
            } elseif ($request->has('image') && $request->input('image') != $item->image) {
                $imageName = $this->storeBase64Image($request);
                $this->removeImageIfExist($item->image);
                $item->image = '/storage/' . $imageName;
            }

            $item->update();

            // Update variants
            $variants = $request->input('variants', []);
            if ($variants && is_array($variants) && count($variants)) {
                // Delete all exist variants if exists
                if ($item->variants()->count()) {
                    $item->variants()->delete();
                }
                $item_variants = (new VariantController())->save_variant($variants, $item->id);
            }

            // Update modifiers
            $modifiers = $request->input('modifiers', []);
            if ($modifiers && is_array($modifiers) && count($modifiers)) {
                $item->modifiers()->sync(array_map(function ($m) {
                    return $m['id'];
                }, $modifiers));
            }

            // Update taxes
            $taxes = $request->input('taxes', []);
            if ($taxes && is_array($taxes) && count($taxes)) {
                $item->taxes()->sync(array_map(function ($tax) {
                    return $tax['id'];
                }, $taxes));
            }

            $price_lists = array_filter($request->input('price_lists'), function ($field) {
                return !!$field['value'];
            });

            if (count($price_lists)) {
                $sync_data = [];
                foreach ($price_lists as $price_list) {
                    $sync_data[$price_list['id']] = ['price' => $price_list['value']];
                }
                $item->price_lists()->sync($sync_data);
            }

            // Department prices are not updated during edit
            $department_prices = array_filter($request->input('departments'), function ($field) {
                return !!$field['selling_price'];
            });

            if (count($department_prices)) {
                foreach ($department_prices as $department_price) {
                    // Add stock item (if doesn't exist) or update department's warehouse
                    $wh_id = Department::find($department_price['id'])->warehouse_id;
                    if ($wh_id) {
                        $wh = Warehouse::find($wh_id)->items();
                        $wh->syncWithoutDetaching($item->id);
                        $wh->updateExistingPivot($item->id, ['selling_price' => $department_price['selling_price']]);
                    }
                }
            }
        });

        return response()->json([
            'status' => 'success',
            'data' => Item::with('user', 'category.department', 'price_lists', 'variants', 'modifiers', 'taxes')->find($id)
        ]);
    }

    public function removeImageIfExist($url)
    {
        if (str_contains($url, "storage")) {
            $path_parts = pathinfo($url);
            Storage::disk('public')->delete(config('app.uploads.item_images') . $path_parts['basename']);
        }
    }

    public function deleteVariant($item_id): JsonResponse
    {
        Variant::where('item_id', $item_id)->delete();
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user', 'category.department', 'price_lists', 'variants', 'modifiers', 'taxes')->find($item_id)
        ]);
    }

    public function changeState($id)
    {
        $item = Item::find($id);
        $item->state = !$item->state;
        $item->save();

        return response()->json([
            'status' => 'success',
            'data' => $item,
        ]);
    }

    public function getArchivedItems()
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user')->where('state', 0)->orderBy('updated_at')->get(),
        ], 200);
    }

    public function massInsert(Request $request)
    {
        $count = null;
        $exists = null;
        DB::transaction(function () use ($request, &$count, &$exists) {
            $user = $request->user();
            $department = Department::where(function ($query) use ($request) {
                $query->whereRaw('LOWER(`name`) like ? ', array(strtolower($request->department)));
            });
            if ($department->exists()) {
                $department = $department->first()->id;
            } else {
                $department = Department::insertGetId(['name' => $request->department, 'user_id' => $user->id]);
            }

            $rows = $request->rows;
            $count = 0;
            foreach ($rows as $k => $row) {

                $category = Category::where(function ($query) use ($row, $k) {
                    $query->whereRaw('LOWER(`name`) like ? ', array(strtolower($k)));
                })->where('department_id', $department);

                if ($category->exists()) {
                    $category = $category->first()->id;
                } else {
                    $category = Category::insertGetId([
                        'name' => ucwords(strtolower($k)),
                        'department_id' => $department,
                        'user_id' => $user->id,
                        'created_at' => date('Y-m-d H:i:s'),
                        'updated_at' => date('Y-m-d H:i:s'),
                    ]);
                }

                $exists = [];

                foreach ($row as $it) {
                    // return [$it['name']];
                    $item = Item::where('category_id', $category)->whereRaw("UPPER(name) LIKE ?", strtoupper($it['name']));
                    if ($item->exists()) {
                        $exists[] = $it;
                    } else {
                        $create = [
                            'code' => get_next_ref('item'),
                            'name' => $it['name'],
                            'unit_price' => $it['selling_price'],
                            'purchase_price' => $it['purchase_price'],
                            'description' => $it['description'],
                            'reorder_level' => $it['reorder_level'],
                            'category_id' => $category,
                            'image' => null,
                            'user_id' => $user->id,
                            'created_at' => date('Y-m-d H:i:s'),
                            'updated_at' => date('Y-m-d H:i:s')
                        ];

                        Item::insert($create);
                        $count++;
                    }

                }

                /*$data = array_map(function ($r) use ($category) {
                    return [
                        'name' => $r['name'],
                        'unit_price' => $r['unit_price'],
                        'category_id' => $category,
                        'image' => 'null',
                        'created_at' => date('Y-m-d H:i:s'),
                        'updated_at' => date('Y-m-d H:i:s')
                    ];
                }, $row);*/
            }
        });


        return response()->json([
            'status' => 'success',
            'created' => $count,
            'exists' => implode(', ', array_map(function ($c) {
                return $c['name'];
            }, $exists)),
            'exists_count' => count($exists)
        ]);
    }


    public function searchItem(Request $request)
    {
        $query = $request->name;
        $items = Item::with('user')
            ->where('name', 'like', '%' . $query . '%')
            ->get();
        $results = [];
        foreach ($items as $item) {
            $data = []; // <= Added
            $data["id"] = $item->id;
            $data["name"] = $item->name;
            $data["unit_price"] = $item->unit_price;
            $data["reorder_level"] = $item->reorder_level;
            $data["optimal_order"] = $item->optimal_order;
            $data["code"] = $item->code;
            $data["barcode"] = $item->barcode;
            // .... add common fields

            $variants = ItemVariant::where('item_id', $item->id)->get();
            if (count($variants)) {
                foreach ($variants as $variant) {
                    $variant["name"] = $data["name"] . " (" . $variant->name . ")";
                    $variant["unit_price"] = $variant->unit_price;
                    $variant["code"] = $variant->code;
                    $variant["barcode"] = $variant->barcode != null ? $variant->barcode : $data["barcode"];
                    $variant["is_variant"] = true;
                    $results[] = $variant;
                }
            } else {
                $data["is_variant"] = false;
                $results[] = $data;
            }
        }
        return response()->json([
            'status' => 'success',
            'data' => $results
        ], 200);
    }

    public function getCategoryItems($category_id)
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user')->where('category_id', $category_id)->get(),
        ], 200);
    }

    public function getSupplierItems($supplier_id)
    {
        return response()->json([
            'status' => 'success',
            'data' => Item::with('user')->where('supplier_id', $supplier_id)->get(),
        ], 200);
    }

}
