<?php

namespace App\Http\Controllers;

use App\RoleUser;
use Auth;
use App\Role;
use App\User;
use App\Permission;
use Illuminate\Http\Request;
use App\Events\RoleSyncable;
use App\Events\UserSyncable;
use Illuminate\Support\Facades\DB;
use Illuminate\Validation\Rule;

class RoleController extends Controller
{
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
    }

    public function index()
    {
        return response()->json(['status' => 'success', 'data' => Role::all()]);
    }

    public function systemRoles()
    {
        return response()->json(['status' => 'success', 'data' => Role::whereSystem(true)->get()]);
    }


    public function store(Request $request)
    {
        $this->validate($request, [
            'name' => 'required|string|max:255',
            'description' => 'required|string'
        ]);

        $role = new Role();
        $role->name = $request->name;
        $role->slug = strtolower($role->name);
        $role->description = $request->description;
        $role->user_id = auth()->id();

        if ($role->save()) {
            return response()->json([
                'status' => 'success',
                'data' => $role
            ]);
        }
        return response()->json([
            'status' => 'error',
            'message' => 'Error while creating role!'
        ]);
    }

    public function update(Request $request, $id)
    {
        $this->validate($request, [
            'description' => 'required|string|max:255',
            'name' => "required|string|unique:roles,name,$id"
        ]);

        $role = Role::find($id);
        $role->name = preg_replace('/\s+/', ' ', trim($request->name));
        $role->description = $request->description;

        // Only change role slug for non-system roles
        if (!$role->system) {
            $role->slug = str_replace(" ", "-", strtolower($role->name));
        }

        if ($role->update()) {
            return response()->json([
                'status' => 'success',
                'data' => $role
            ]);
        }
        return response()->json([
            'status' => 'error',
            'message' => 'Error while updating role!',
            'data' => $role
        ]);
    }

    public function delete($id)
    {
        $role = Role::find($id);

        if ($role->users()->count()) {
            return response()->json([
                'status' => 'error',
                'message' => 'Role cannot be deleted because it is current attached to a user.'
            ]);
        }
        // @todo Deleting a system role
        if ($role->system) {
            return response()->json([
                'status' => 'error',
                'message' => 'System roles cannot be deleted.'
            ]);
        }
        if ($role->delete()) {
            return response()->json([
                'status' => 'success'
            ]);
        } else {
            return response()->json([
                'status' => 'error',
                'message' => 'Role record not found!'
            ]);
        }
    }

    public function userRoles($id)
    {
//        return response()->json(['data'=>User::find($id)->roles]);
        return response()->json([
            'status' => 'success',
            'data' => User::find($id)->roles
        ]);
    }

    public function authRolePermissions()
    {
        return response()->json([
            'status' => 'success',
            'data' => auth()->user()->roles()->with('permissions')->get()
        ]);
    }

    public function userPermissions()
    {
        $roles = User::find(auth()->id())->roles()->with('permissions')->get();
        $data = [];
        foreach ($roles as $role) {
            $data[] = $role->permissions->pluck('slug');
        }
        // return response()->json(['data' => $data]);
        return response()->json($data);
    }

    public function permissions()
    {
        return response()->json([
            'status' => 'success',
            'data' => Permission::all()
        ]);
    }

    public function rolePermissions($id)
    {
//        dd(Role::find($id)->permissions);
        return response()->json([
            'status' => 'success',
            'data' => Role::find($id)->permissions
        ]);
    }

    public function assignUserRoles(Request $request, $id)
    {
        $user = User::find($id);
        $user->roles()->sync($request->roles);
        return response()->json([
            'status' => 'success'
        ]);
    }

    public function assignRolePermissions(Request $request, $id)
    {
        $role = Role::find($id);
        $role->permissions()->sync($request->permissions);
        return response()->json([
            'status' => 'success',
            'message' => 'Permissions assigned'
        ]);
    }

}
