<?php

namespace App\Http\Controllers;

use App\Customer;
use App\Item;
use App\SalesOrder;
use App\SalesOrderItem;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Schema;

class SalesOrderController extends Controller
{
    private const STATUS_COMPLETE = '1';
    private const STATUS_EXPIRED = '-1';

    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
    }

    public function index()
    {
        $sale_order = SalesOrder::with(['user', 'items', 'service'])->latest()->limit(100)->get();
        return response()->json([
            'status' => 200,
            'data' => $sale_order,
        ], 200);

    }


    /**
     * Get all orders within a session
     * Called from within POS (service can see all orders within a session)
     */
    public function getSessionSalesOrder()
    {
        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::ownerOrders()->pendingOrders()->with(['items', 'service', 'user'])->get()
        ]);
    }

    /**
     * Get all orders within a session pertaining to a unique service (user)
     * Called from mobile service-assist application
     */
    public function getServiceOrders()
    {
        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::ownerOrders()->pendingOrders()->with(['items', 'service', 'user'])->get()
        ]);
    }

    public function getSaleOrder($id)
    {
        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::with(['user', 'items', 'service', 'customer'])->find($id)
        ], 200);
    }

    public function _register_sales_order(Request $request, $id = null)
    {
        DB::beginTransaction();
        $user = auth()->user();
        $cart_items = $request->cart;
        $table_name = app(SalesOrder::class)->getTable();

        if (is_null($id)) {
            $sales_order = new SalesOrder();
            // values at set only once
            $sales_order->code = get_next_ref('sales_order');
            // $sales_order->session_id = $user->session->id;

            if (Schema::hasColumn($table_name, 'department_id')) {
                $sales_order->department_id = $user->department_id;
            }

        } else {
            $sales_order = SalesOrder::find($id);
            // Delete all items and reattach
            $sales_order->items()->delete();
        }

        $sales_order->discount = $request->discount;
        $sales_order->customer_id = $request->input('customer');

        if (Schema::hasColumn($table_name, 'service_id')) {
            $sales_order->service_id = $request->input('service'); // @todo set service as auth-user when request comes from mobile assist
        }
        if (Schema::hasColumn($table_name, 'owner_id')) {
            $sales_order->label = $request->input('label', ('SOR - ' . date('h:i A')));
            $sales_order->owner_id = auth()->id();
            $sales_order->creator_id = auth()->id();
            $sales_order->comment = $request->input('comment');

        } else {
            // Catch breaking changes
            $sales_order->user_id = auth()->id();
            $sales_order->service_id = $request->input('service');
        }
        if ($id) {
            $sales_order->update();
        } else {
            $sales_order->save();
        }

        $items = [];
        $sum = 0;
        foreach ($cart_items as $item) {
            $real_item = auth()->user()->department->warehouse->items()->where('item_id', $item['id'])->with('price_lists')->first();
            $unit_price = null;
            if ($request->input('customer', false)) { // @todo Ensure the correct unit-price is used (preference: customer (price-list) > department > article).
                $customer = Customer::find($request->input('customer'));
                $pls = array_map(function ($pl) {
                    return ['id' => $pl['id'], 'price' => $pl['pivot']['price'], 'name' => $pl['name']];
                }, $real_item->price_lists->toArray());

                $index = array_search($customer->price_list_id, array_column($pls, "id"));
                if ($index >= 0) {
                    $unit_price = $pls[$index]['price'];
                }
            }
            if ($unit_price <= 0) {
                $unit_price = $real_item->pivot->selling_price ?: $real_item['unit_price'];
            }


            $sum += $item['quantity'] * $unit_price;

            $items[] = new SalesOrderItem([
                'item_id' => $item['id'],
                'quantity' => $item['quantity'],
                'unit_price' => $unit_price,
                // 'session_id' => $sales_order->session_id,
            ]);
        }
        $sales_order->items()->saveMany($items);

        if (Schema::hasColumn($table_name, 'total')) {
            $sales_order->total = $sum;
            $sales_order->update();
        }

        DB::commit();
        return $sales_order->id;
    }

    public function create(Request $request)
    {
        $inserted_id = $this->_register_sales_order($request);
        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::with('items', 'service', 'user')->find($inserted_id)
        ]);
    }

    public function delete($id)
    {
        DB::beginTransaction();
        $sale_order = SalesOrder::find($id);
        $sale_order->delete();
        DB::commit();
        return response()->json([
            'status' => 'success'
        ]);
    }

    public function update(Request $request, $id)
    {
        $inserted_id = $this->_register_sales_order($request, $id);
        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::with('items', 'user')->find($inserted_id)
        ]);
    }

    public function updateTicketLabel(Request $request, $id)
    {
        $order = SalesOrder::find($id);
        if ($order) {
            $order->label = $request->input('label', $order->label);
            $order->update();
        }

        return response()->json(['status' => 'success', 'data' => $order]);
    }

    public function forceDelete($id): JsonResponse
    {
        DB::beginTransaction();
        $sale_order = SalesOrder::withTrashed()->find($id);
        $sale_order->forceDelete();

        DB::commit();
        return response()->json([
            'status' => 'success'
        ]);
    }

    public function markComplete($id)
    {
        $sale_order = SalesOrder::find($id);
        if ($sale_order) {
            $sale_order->status = self::STATUS_COMPLETE;
            $sale_order->update();
        }
    }

    public function revoke($id)
    {
        $sale_order = SalesOrder::find($id);
        if ($sale_order) {
            $sale_order->status = self::STATUS_EXPIRED;
            $sale_order->update();
        }
    }

    public function deleteOrders(Request $request): JsonResponse
    {
        $orders = SalesOrder::whereIn('id', $request->input('orders'));
        $orders->delete();
        return response()->json([
            'status' => 'success'
        ]);
    }

    public function mergeOrders(Request $request): JsonResponse
    {
        $order = SalesOrder::find($request->input('order'));
        $other_orders = SalesOrder::whereIn('id', $request->input('others'));
        $order_item_ids = array_map(function ($item) {
            return $item['item_id'];
        }, $order->items->toArray()); // array_merge();
        foreach ($other_orders->get() as $other_order) {
            foreach ($other_order->items as $item) {
                if (in_array($item->item_id, $order_item_ids)) {
                    SalesOrderItem::where([['sales_order_id', '=', $order->id], ['item_id', '=', $item['item_id']]])->increment('quantity', $item->quantity);
                    SalesOrderItem::where([['sales_order_id', '=', $other_order->id], ['item_id', '=', $item['item_id']]])->delete();
                } else {
                    SalesOrderItem::where([['sales_order_id', '=', $other_order->id], ['item_id', '=', $item['item_id']]])->update(['sales_order_id' => $order->id]);
                }
            }
            $other_order->update(['comment' => 'MERGED']);
        }

        $order->label = $order->label . '-MERGED';
        $order->update();

        $other_orders->delete();

        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::with(['user', 'items'])->find($order->id)
        ]);
    }

    public function splitOrder(Request $request): JsonResponse
    {
        $order = SalesOrder::find($request->input('order'));
        $other_order_item_ids = $request->input('items');

        $new_order = $order->replicate();
        $new_order->label = ($new_order->label . '-1');
        $new_order->code = get_next_ref('sales_order');
        $new_order->save();

        SalesOrderItem::where('sales_order_id', $order->id)->whereIn('item_id', $other_order_item_ids)->update(['sales_order_id' => $new_order->id]);

        return response()->json([
            'status' => 'success',
            'data' => SalesOrder::with(['items', 'user'])->whereIn('id', [$order->id, $new_order->id])->get()
        ]);
    }

    public function transferOrders(Request $request): JsonResponse
    {
        $query = SalesOrder::with(['items', 'user'])->whereIn('id', $request->input('orders', []));
        $query->update(['owner_id' => $request->input('new_owner', auth()->id())]); // fallback to previous owner

        return response()->json([
            'status' => 'success',
            'data' => $query->get()
        ]);
    }

}
