<?php

namespace App\Http\Controllers;

use App\Config;
use App\Invoice;
use App\Role;
use App\Sale;
use PDF;
use App\User;
use App\Item;
use App\Session;
use App\Department;
use Carbon\Carbon;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Log;
use Illuminate\Validation\Rule;
use Illuminate\Validation\ValidationException;

class SessionController extends Controller
{
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
    }

    public function index(Request $request): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Session::with('user')->latest()->get()
        ]);
    }

    public function getSession($id): array
    {
        return [
            'status' => 'success',
            'data' => Session::whereId($id)->with(['user', 'department', 'opener', 'closer'])->first()];
    }

    public function getSessionableUsers(): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Role::whereSlug('cashier')->whereSystem(true)->first()->users
        ]);
    }

    public function getDepartmentSessions($id, Request $request): JsonResponse
    {
        $sessions = Session::with(['user', 'department', 'opener', 'closer'])
            ->where('department_id', $id)->latest()->paginate($request->input('perPage', 9));

        return response()->json([
            'status' => 'success',
            'data' => $sessions
        ]);
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param Request $request
     * @return JsonResponse
     * @throws ValidationException
     */
    public function store(Request $request): JsonResponse
    {
        $this->validate($request, [
            'user' => 'required',
            'department' => 'required'
        ]);
        $user = User::find($request->input('user'));
        $user_has_session = Session::whereUserId($user->id)->whereStatus(true)->exists();
        if ($user_has_session) {
            return response()->json([
                'status' => 'error'
            ], 402);
        }

        $session = new Session();
        $session->user_id = $request->input('user');
        $session->department_id = $request->input('department');
        $session->opener_id = auth()->id();

        $session->save();
        return response()->json([
            'status' => 'success',
            'data' => Session::with('user', 'department', 'opener', 'closer')->find($session->id)
        ]);
    }

    /**
     * Close session by logged-in cashier.
     *
     * @return JsonResponse
     */
    public function cashier_close_session(): JsonResponse
    {
        $session = $this->_close_session_internal(request()->user()->session->id);
        return response()->json([
            'status' => 'success',
            'data' => $session
        ]);
    }

    /**
     * Close session by $id
     * @param int $id
     * @return JsonResponse
     */
    public function close(int $id): JsonResponse
    {
        $this->_close_session_internal($id);
        return response()->json([
            'status' => 'success',
            'data' => Session::with('user', 'department', 'opener', 'closer')->find($id)
        ]);
    }

    public function _close_session_internal($id)
    {
        $session = Session::find($id);
        if ($session) {
            $session->closed_on = date('Y-m-d H:i:s');
            $session->closer_id = auth()->id();
            $session->status = false;
            $session->update();
        }
        return $session;
    }

    /**
     * @throws ValidationException
     */
    public function generateSessionReport(Request $request): JsonResponse
    {
        $this->validate($request, [
            'sessions' => 'required|array|min:1'
        ]);

        $sessions = Session::with(['user' => function ($query) {
            $query->select('id', 'name');
        }])->whereIn('id', $request->sessions)->get();

        $names = "";
        $dates = [];
        foreach ($sessions as $index => $session) {
            if (!count($dates)) {
                $dates[0] = $session->closed_on;
                $dates[1] = count($sessions) > 1 ? $session->closed_on : null;
            }
            if (count($sessions) > 1 && $session->closed_on->lt($dates[0])) {
                $dates[0] = $session->closed_on;
            }
            if (count($sessions) > 1 && $session->closed_on->gt($dates[1])) {
                $dates[1] = $session->closed_on;
            }
            if (!str_contains($names, $session->user->name)) {
                if ($index > 0) $names .= ", ";
                $names .= $session->user->name;
            }
        }

        $sales = Sale::with(['items.item.category', 'service', 'payments'])->whereIn('session_id', $request->sessions);
        $discount = $sales->sum('discount') ?: 0;

        $payment_methods = [];
        $item_ids = [];
        foreach ($sales->get() as $sale) {
            foreach ($sale->payments as $payment) {
                if (!isset($payment_methods[$payment->name])) {
                    $payment_methods[$payment->name] = 0;
                }
                $payment_methods[$payment->name] += $payment->pivot->amount;
            }
            $item_ids = array_merge($item_ids, $sale->items->pluck('item_id')->toArray());
        }
        $item_ids = array_unique($item_ids);

        $items = Item::whereIn('id', $item_ids)->select('id', 'name', 'category_id')
            ->with(['category' => function ($q) {
                return $q->select('id', 'name');
            }])->get()->toArray();
        $cat_items = [];
        array_map(function ($item) use (&$cat_items, $request) {
            if (!isset($cat_items[$item['category']['name']])) {
                $cat_items[$item['category']['name']] = [];
            }
            $sub_total = 0;
            $sub_data = [
                'name' => '',
                'qty' => 0,
                'unit_price' => 0,
                'total' => 0
            ];
            $invoices = Invoice::whereIn('session_id', $request->sessions)->whereItemId($item['id'])->get();
            foreach ($invoices as $idx => $inv) {
                if (!$idx) { // Get for first inv only (index 0)
                    $sub_data['name'] = $item['name'];
                    $sub_data['unit_price'] = $inv->unit_price;
                }
                $sub_data['qty'] += (int)$inv->quantity;
                $sub_data['total'] += (int)$inv->unit_price * (int)$inv->quantity;
            }
            $cat_items[$item['category']['name']][] = $sub_data;
        }, $items);

        $start_date = Carbon::parse($dates[0])->toDayDateTimeString();
        $end_date = !$dates[1] ? null : Carbon::parse($dates[1])->toDayDateTimeString();

        $file_start = (date('Y-m-d His'));

        Log::debug(json_encode($cat_items));

        $pdf = PDF::loadView('reports/session_report', [
            'cat_items' => $cat_items, 'discount' => $discount,
            'payment_methods' => $payment_methods,
            'start_date' => $start_date, 'end_date' => $end_date,
            'names' => $names
        ]);
        $path = storage_path('app/public/');
        $filename = 'tmp/session_report_' . $file_start . '.pdf';
        $pdf->save($path . $filename);

        return response()->json(['status' => 'success', 'url' => url('storage/' . $filename)]);
    }

}
