<?php

namespace App\Http\Controllers;

use App\Account;
use App\AccountCategory;
use App\DefaultMessage;
use App\ItemHistory;
use App\Supplier;
use App\Supply;
use App\Warehouse;
use App\Withdrawal;
use App\Item;
use Illuminate\Http\Request;
use App\Http\Controllers\WarehouseController;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Schema;

class SupplyController extends Controller
{
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
        //
    }

    public function getSupplies()
    {
        return response()->json([
            'status' => 'success',
            'data' => Supply::with(['supplier' => function ($query) {
                return $query->select('id', 'name', 'phone', 'address');
            }, 'items.item', 'user' => function ($query) {
                return $query->select('id', 'name', 'username');
            }])->withCount('items')->latest()->get()
        ]);
    }

    public function getSupply($id)
    {
        return response()->json([
            'status' => 'success',
            'data' => Supply::find($id)
        ], 200);
    }

    public function createSupplies(Request $request)
    {
        $this->validate($request, [
            'supplier' => 'required|exists:suppliers,id',
            'paid' => 'required',
            'date' => 'required|date',
            'items' => 'required|min:1',
            'warehouse' => 'required'
        ]);

        DB::beginTransaction();
        $items = $request->input('items');

        // Summing up all the prices of that particular purchase
        $total_price = array_sum(array_map(function ($item) {
            return $item['quantity'] * $item['unit_price'];
        }, $items));

        $amt_owed = $total_price - $request->input('paid');
        $amt_paid = $request->input('paid');

        if ($amt_paid > $total_price) {
            DB::rollBack();
            return response()->json([
                'status' => 'error',
                'message' => 'Invalid transaction! Please ensure that the amount paid is not greater than total price'
            ], 403);
        }
        // 1. find subcategory
        $category = AccountCategory::where('slug', 'like', '%payment%')->first();
        if ($category) {
            $subcategory = $category->subcategories()->where('slug', 'like', '%supplier%')->first();
            if (!$subcategory) {
                DB::rollBack();
                return response()->json([
                    'status' => 'error',
                    'message' => "A category exists called $category->name.
                    But there is no subcategory called *supplier*.
                    Please, create one before effecting this purchase."
                ]);
            }
        } else {
            DB::rollBack();
            return response()->json([
                'status' => 'error',
                'message' => 'Ensure that there exist an account category called Payment. And a subcategory called *supplier*.
                              This is required to register the supplier payment for this purchase.'
            ]);
        }

        $supply = new Supply();
        $supply->supplier_id = $request->input('supplier');
        $supply->user_id = auth()->id();
        $supply->total_cost = $total_price;
        $supply->paid = $amt_paid;
        $supply->owed = $amt_owed;

        if (Schema::hasColumn('supplies', 'code')) {
            $supply->code = get_next_ref('purchase');
        }

        if (Schema::hasColumn('supplies', 'date')) {
            $supply->date = $request->input('date');
        }

        $supply->save();
        $whCtrl = new WarehouseController();
        $default_wh = Warehouse::whereDefault(true)->first();
        $wh_id = $request->input('warehouse');

        if ($default_wh) {
            // Transfer to default WH
            $whCtrl->transferArticle($items, $default_wh->id, null);

            if ($default_wh->id != $wh_id) {
                $whCtrl->transferArticle($items, $wh_id, $default_wh->id, null);
            }
        }

        $supply->items()->createMany(
            array_map(function ($item) use ($supply, $request) {
                $message = DefaultMessage::where('module', 'item_purchase')->first();
                $supplier = Supplier::find($supply->supplier_id);
                $item_rec = Item::find($item['id']);
                $req = new Request();
                $req->setMethod('POST');
                $req->request->add([
                    'item_id' => $item['id'],
                    'in' => $item['quantity'],
                    'type' => trans('supply.item_history.type.goods_in'),
                    'date' => $request->input('date'),
                    'reference' => $supply->code,
                    'details' => str_replace(
                        ['{supplier}', '{item}', '{item_ref}', '{inv_ref}', '{quantity}', '{date}', '{created}'],
                        [$supplier->name, $item_rec->name, $item_rec->code, $supply->code, $item['quantity'], $supply->date, substr($supply->created_at, 0, 10)], $message->message),
                    'cost_price' => $item['unit_price'],
                ]);
                $historyCtrl = new ItemHistoryController();
                $historyCtrl->store($req);

                // Calculate average price of item if necessary
                $this->_standardize_cost_price($item['id'], $item['unit_price']);

                return [
                    /*'supply' => $supply->id,*/
                    'item_id' => $item['id'],
                    'unit_price' => $item['unit_price'],
                    'quantity' => $item['quantity'],
                    'total_cost' => $item['quantity'] * $item['unit_price']
                ];
            }, $items)
        );

        $supplier = Supplier::whereId($supply->supplier_id);
        if ($amt_owed > 0) {
            $supplier->increment('amt_owed', $amt_owed);
        }
        $supply->supplier->payments()->create([
            'amount' => $amt_paid,
            'user_id' => auth()->id(),
            'message' => __('Payment after supply of items')
        ]);

        // 2. increment subcategory amount by $amt_paid
        // 3. update account's amount by $amt_paid
        // 4. create withdrawal with $amt_paid

        $subcategory->increment('amount', $amt_paid);
        // Create a new withdrawal depending on the amount paid from the main account

        $withdraw = new WithdrawalController();
        $req = $request;
        $req->request->add([
            'message' => ("Supplier payment: " . $supplier->first()->name),
            'subcategory_id' => $subcategory->id,
            'date' => date('Y-m-d'),
            'amount' => $amt_paid
        ]);
        $withdraw->makeWithdrawal($req);

        $account = Account::whereStatus(true)->latest()->first();
        $account->amount -= $amt_paid;
        $account->update();

        /*
         * @todo Add purchased items to stock
         * @todo If department is passed, then transfer items to department immediately
         * */

        // Save new products into the warehouse
        /*foreach ($items as $item) {
            $warehouseCtrl = new WarehouseController();
            $req = $request;

            $req->request->add(['quantity' => $item['quantity']]);
            $warehouseCtrl->updateItem($item['id'], $request);

            if ($request->input('department_id')) {
                $stockCtrl = new StockController();
                $stockCtrl->updateStock($item, $request->input('department_id'));
            }
        }*/

        //  Move items directly to a particular department id department_id is specified
        //  if ($request->input('department_id')) {
        //      $stockCtrl = new StockController();
        //      $stockCtrl->updateStock($items, $request->input('department_id'));
        //  }
        DB::commit();
        return response()->json([
            'status' => 'success',
            'data' => $supply
        ], 200);

    }

    public function _standardize_cost_price ($id, $new_price)
    {
        $item = Item::find($id);
        if ($item->purchase_price != $new_price && $new_price > 0) {
            $item->purchase_price = $item->purchase_price ? ($item->purchase_price + $new_price)/2 : $new_price;
            $item->update();
        }
        return $item;
    }

    public function updateSupplies($id, Request $request)
    {
        $this->validate($request, [
            'supplier_id' => 'required',
            'user_id' => 'required',
            'total_cost' => 'required',
            'paid' => 'required',
        ]);

        $supplies = Supplier::find($id);
        $supplies->total_cost = $request->input('total_cost');
        $supplies->paid = $request->input('paid');

        $supplies->update();
        return response()->json([
            'status' => 'success',
            'data' => $supplies
        ], 200);
    }
}
