<?php

namespace App\Http\Controllers;

use App\Invoice;
use App\Item;
use App\ItemHistory;
use App\ItemSupply;
use App\Stock;
use App\StockAdjustment;
use App\Transfer;
use App\Warehouse;
use Illuminate\Database\Eloquent\Relations\BelongsToMany;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Validation\ValidationException;

class WarehouseController extends Controller
{
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
        //
    }

    public function fetchWarehouses()
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with('user')->get()
        ]);
    }

    public function all()
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with(['user'])->get()
        ], 200);
    }

    public function getWarehouse($id): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::find($id)
        ]);
    }

    public function getWarehouseItems($id): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with('items.category')->find($id)
        ]);
    }

    // Get Warehouse Items only
    public function getItems($id): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with('items')->find($id)->items
        ]);
    }

    public function getWarehousesItem($id): JsonResponse
    {
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with(['items' => function (BelongsToMany $query) use ($id) {
                $query->where('item_id', $id);
            }])->get()
        ]);
    }


    /**
     * @throws ValidationException
     * @todo Add warehouse display name (name) and a unique immutable code for reference
     */
    public function createWarehouse(Request $request): JsonResponse
    {
        $this->validate($request, [
            'name' => 'required|unique:warehouses',
            'description' => 'nullable'
        ]);

        $wh = new Warehouse();
        $wh->name = $request->input('name');
        $wh->description = $request->input('description');
        $wh->user_id = auth()->id();

        $wh->save();
        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with(['user'])->find($wh->id)
        ]);
    }

    public function updateWarehouse($id, Request $request)
    {
        $this->validate($request, [
            'name' => "required|unique:warehouses,name,$id",
            'description' => 'nullable'
        ]);
        $wh = Warehouse::find($id);
        $wh->name = $request->input('name');
        $wh->description = $request->input('description');
        $wh->update();

        return response()->json([
            'status' => 'success',
            'data' => Warehouse::with(['user'])->find($wh->id)
        ], 200);
    }

    public function changeDefaultWarehouse($id)
    {
        Warehouse::whereDefault(true)->decrement('default');
        Warehouse::whereId($id)->increment('default');

        return response()->json([
            'status' => 'success'
        ], 200);
    }

    public function transferArticle($items, $to_id, $from_id = null, $comment = null)
    {
        if (is_int($items)) $items = [['id' => $items, 'quantity' => 0]]; // If integer is passed
        if (is_array($items) && isset($items[0]) && is_int($items[0])) { // If integer array is passed
            $data = [];
            foreach ($items as $item) {
                $data[] = ['id' => $item, 'quantity' => 0];
            }
            $items = $data;
        }

        if (is_array($items)) { // items must be array of array(s)
            DB::transaction(function () use ($items, $from_id, $to_id, $comment) {
                /*$default_wh = Warehouse::whereDefault(true)->first();
                if (!$from_id && $to_id != $default_wh->id && $move_to_default) {
                    $from_id = $default_wh->id;
                }*/
                foreach ($items as $item) {
                    if ($from_id) {
                        $stock = Stock::where([['item_id', '=', $item['id']], ['warehouse_id', '=', $from_id]])->first();

                        if ($stock) {
                            if ($stock->quantity >= $item['quantity']) {
                                $this->_update_warehouse($item, $from_id, true);
                            }
                        } elseif (!$item['quantity']) { // When quantity is zero
                            $this->_update_warehouse($item, $from_id);
                        }
                    }
                    // Execute to transfer
                    if ($to_id) {
                        $this->_update_warehouse($item, $to_id);
                    }
                }
                $this->_register_transfer($items, $from_id, $to_id, $comment);
            });
        }
    }

    private function _update_warehouse($item, $wh_id, $reduce = false)
    {
        if (!is_array($item) && is_int($item)) {
            $item = ['id' => $item];
        }
        $search = Stock::where([['item_id', $item['id']], ['warehouse_id', $wh_id]]);
        $wh = Warehouse::find($wh_id);
        $real_item = Item::find($item['id']);
        if (!array_key_exists('selling_price', $item)) {
            $item['selling_price'] = $real_item['unit_price'];
        }
        if (!array_key_exists('quantity', $item)) {
            $item['quantity'] = 0;
        }

        if ($search->exists()) {
            $search = $search->first();
            $wh->items()->updateExistingPivot($item['id'], [
                'quantity' => $search->quantity + ($reduce ? -$item['quantity'] : $item['quantity'])
            ]);
        } else {
            $wh->items()->attach($item['id'], [
                'selling_price' => $item['selling_price'],
                'quantity' => $item['quantity']
            ]);
        }
    }

    private function _register_transfer($items, $from_id, $to_id, $comment)
    {
        if (!is_array($items)) return;
        if (!is_array($items[0])) $items = [$items];
        Transfer::insert(array_map(function ($item) use ($from_id, $to_id, $comment) {
            return [
                'item_id' => $item['id'],
                'quantity' => $item['quantity'],
                'from_id' => $from_id,
                'to_id' => $to_id,
                'comment' => $comment,
                'user_id' => auth()->id(),
                'created_at' => date('Y-m-d H:i:s'),
                'updated_at' => date('Y-m-d H:i:s')
            ];
        }, $items));
    }


    public function updateWarehouseItem($id, Request $request)
    {
        $this->validate($request, [
            'quantity' => 'required',
        ], [
            'quantity.required' => 'Please Provide the item quantity.',
        ]);
        $wh = Warehouse::where('item_id', $id);
        if ($wh->exists()) {
            $wh->increment('quantity', $request->input('quantity'));
            $wh = $wh->first();

        } else {
            $wh = new Warehouse();
            $wh->item_id = $id;
            $wh->quantity = $request->input('quantity');
            $wh->unit_type = $request->input('unit_type') || Item::find($id)->unit_type;

            $wh->save();
        }

        return response()->json([
            'status' => 'success',
            'data' => $wh
        ], 200);
    }


    public function updateItemSellingPrice($id, $item_id, Request $request)
    {
        Warehouse::find($id)->items()
            ->updateExistingPivot($item_id, ['selling_price' => $request->input('price')]);

        return response()->json([
            'status' => 'success'
        ]);
    }

    public function getStockItem($id): JsonResponse
    {
        $item = Item::with('warehouses')->find($id);
        $last_sale = Invoice::where('item_id', $id)->select('created_at')->latest()->first();
        $last_purchase = ItemSupply::with('supply')->where('item_id', $id)->latest()->first();
        /*Warehouse::with(['items' => function ($query) use ($id) {
            return $query->where('item_id', $id);
        }])->get();*/
        return response()->json([
            'status' => 'success',
            'data' => [
                'item' => $item,
                'last_sale' => $last_sale,
                'last_purchase' => $last_purchase ? $last_purchase->supply->date : null,
                'history' => ItemHistory::with('user')->where('item_id', $id)->latest()->get()
            ]
        ]);
    }

    /**
     * @throws ValidationException
     */
    public function adjust_warehouse_stock($warehouse, Request $request): JsonResponse
    {
        /**
         * Procedure:
         * 1. StockAdjustment Model (id, reason, warehouse_id, date, created_at, updated_at)
         * 2. StockAdjustmentItem (item_id, stock_adjustment_id, from, to, unit_price, avr_purchase_price)
         * 3. Migrate all previous adjustment (registered in item_history table) to new structure
         * 4. @todo register Purchase_Price_History changes.
         * 5. @todo Check possible of having an item without a purchase price.
         * 6. @todo Update stock values,
         * 7. @todo Create invoice to be validated.
         * 8. @todo Add expiry date to articles, automate message alert
         * 9. @todo Generate barcodes for articles
         */
        $this->validate($request, [
            'items' => 'required|array|min:1',
            'date' => 'required|date',
            'reason' => 'required'
        ]);
        $adjustment = new StockAdjustment();

        DB::transaction(function () use ($request, &$adjustment, $warehouse) {

            $items = $request->input('items');

            $adjustment->reason = $request->input('reason');
            $adjustment->date = $request->input('date');
            $adjustment->warehouse_id = $request->input('warehouse_id', $warehouse);
            $adjustment->user_id = auth()->id();
            $adjustment->save();

            $adjustment->items()->createMany(array_map(function ($item) use ($adjustment) {
                return [
                    'item_id' => $item['id'],
                    'stock_adjustment_id' => $adjustment->id,
                    'from' => $item['pivot']['quantity'],
                    'to' => $item['new_qty'],
                    'unit_price' => $item['unit_price'],
                    'avr_purchase_price' => $item['purchase_price'],
                    'created_at' => date('Y-m-d H:i:s'),
                    'updated_at' => date('Y-m-d H:i:s')
                ];
            }, $items));
        });

        return response()->json([
            'status' => $adjustment->id ? 200 : 403,
            'data' => $adjustment ?? null
        ]);
    }

    public function get_stock_history(Request $request): JsonResponse
    {
        $data = StockAdjustment::with('user', 'warehouse');
        if ($request->has('warehouses')) {
            $data = $data->whereIn('warehouse_id', $request->input('warehouses'));
        }
        if ($request->has('users')) {
            $data = $data->whereIn('user_id', $request->input('users'));
        }
        if ($request->has('date')) {
            $data = $data->whereDate('date', $request->input('date'));
        }
        if ($request->has('items')) {
            $data = $data->whereHas('items', function ($query) use ($request) {
                $query->whereIn('item_id', $request->input('items'))->whereColumn('from', '!=', 'to');
            });
        }

        $data = $data->latest()->get();

        return response()->json([
            'status' => 200,
            'data' => $data
        ]);
    }

    public function get_stock_history_items($id): JsonResponse
    {
        $data = StockAdjustment::with('items.item')->find($id);
        return response()->json([
            'status' => 200,
            'id' => $id,
            'data' => $data
        ]);
    }

    /**
     * @throws ValidationException
     */
    public function adjust_stock($id, Request $request): JsonResponse
    {
        $this->validate($request, [
            'quantity' => 'required|min:1',
            'reason' => 'required',
            'date' => 'required'
        ]);

        $quantity = $request->quantity;
        $type = $request->type;
        $reason = $request->reason;
        $date = $request->date;
        $cost_price = $request->cost_price;
        $sales_price = $request->sales_price;
        $warehouse = $request->warehouse;
        $ref = $request->ref ?: ("@" . auth()->user()->username); // if reference isn't specified, username is used.

        DB::beginTransaction();

        // Update warehouse with stock level
        $this->_update_warehouse(['id' => $id, 'quantity' => $quantity], $warehouse, str_starts_with($type, 'de'));
        if (str_starts_with($type, 'in') && $cost_price) {
            $supplyCtrl = new SupplyController();
            $supplyCtrl->_standardize_cost_price($id, $cost_price);
        }

        // Add stock adjustment to history
        $req = new Request();
        $req->setMethod('POST');
        $req->request->add([
            'item_id' => $id,
            (str_starts_with($type, 'in') ? 'in' : 'out') => $quantity,
            'type' => trans('supply.item_history.type.adjust_' . (str_starts_with($type, 'in') ? 'in' : 'out')),
            'date' => $date,
            'details' => $reason,
            'reference' => $ref,
            'cost_price' => str_starts_with($type, 'in') ? $cost_price : null,
            'sales_price' => str_starts_with($type, 'de') ? $sales_price : null
        ]);
        $historyCtrl = new ItemHistoryController();
        $historyCtrl->store($req);

        DB::commit();

        return response()->json([
            'status' => 'success'
        ]);
    }

}
