<?php

namespace App;

use Illuminate\Auth\Authenticatable;
use Illuminate\Contracts\Auth\Access\Authorizable as AuthorizableContract;
use Illuminate\Contracts\Auth\Authenticatable as AuthenticatableContract;
use Illuminate\Database\Eloquent\Model;
use Laravel\Lumen\Auth\Authorizable;
use Tymon\JWTAuth\Contracts\JWTSubject;

class User extends Model implements AuthenticatableContract, AuthorizableContract, JWTSubject
{
    use Authenticatable, Authorizable;

    protected $primaryKey = 'id';

    /**
     * The attributes that are mass assignable.
     *
     * @var array
     */
    protected $fillable = [
        'name', 'email',
    ];

    /**
     * The attributes excluded from the model's JSON form.
     *
     * @var array
     */
    protected $hidden = [
        'password',
    ];

    /**
     * Get the identifier that will be stored in the subject claim of the JWT.
     *
     * @return mixed
     */
    public function getJWTIdentifier()
    {
        return $this->getKey();
    }

    /**
     * Return a key value array, containing any custom claims to be added to the JWT.
     *
     * @return array
     */
    public function getJWTCustomClaims()
    {
        return [];
    }

    public function getPhotoAttribute($url): string
    {
        return url($url);
    }

    public function getActiveAttribute($active): bool
    {
        return (bool)$active;
    }

    public function getIsAdminAttribute($is_admin): bool
    {
        return (bool)$is_admin;
    }

    public function getConfig() {
//        return Config::latest()->first();
        return \App\Config::latest()->first();
    }

    public function roles()
    {
        return $this->belongsToMany(Role::class);
    }

    public function department()
    {
        return $this->belongsTo(Department::class);
    }

    public function session()
    {
        return $this->hasOne(Session::class)->where('status', 1);
    }

    private function getPermissions()
    {
        $roles = $this->roles()->with('permissions')->get();

        $userPermissions = [];
        foreach ($roles as $role) {

            foreach ($role->permissions->pluck('slug') as $perm) {
                if (!in_array($perm, $userPermissions)) {
                    $userPermissions[] = $perm;
                }
            }
        }
        return $userPermissions;
    }

    public function hasRole($role) {
        return in_array($role, array_map(function ($r) use($role) {return $r['slug'];}, $this->roles->toArray()));
    }

    public function hasAccess(array $permissions): bool
    {
        $userPermissions = $this->getPermissions();
        return (count(array_diff($userPermissions, $permissions)) + count($permissions) == count($userPermissions));
    }

    public function hasAtLeastAccess(array $permissions)
    {
        $userPermissions = $this->getPermissions();
        foreach ($permissions as $permission) {
            if (in_array($permission, $userPermissions))
                return true;
        }
        return false;
    }

    public function user()
    {
        return $this->belongsTo(User::class, 'creator_id', 'id');
    }

    public function scopeActive($query)
    {
        return $query->where('active', true);
    }

    public function scopeServices($query)
    {
        return $query->whereHas('roles', function ($query) {
            $query->where('slug', 'service');
        });
    }

    public function scopeNonServices($query)
    {
        return $query->whereDoesntHave('roles', function ($query) {
            $query->where('slug', 'service');
        });
    }

    public function scopeCashiers($query)
    {
        return $query->whereHas('roles', function ($query) {
            $query->where('slug', 'cashier');
        });
    }

    public function sales_orders()
    {
        return $this->hasMany(SalesOrder::class, 'owner_id', 'id');
    }

    public function sales_orders_departments()
    {
        return $this->hasMany(SalesOrder::class, 'department_id', 'department_id');
    }


}
